@@ -1,3 +1,4 @@
+/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /*
  * src/manager.c - GNOME Volume Manager
  *
@@ -25,6 +26,17 @@
 #include <signal.h>
 #include <unistd.h>
 
+#include <locale.h>
+
+#ifdef HAVE_CODESET
+#include <langinfo.h>
+#endif
+
+#ifdef ENABLE_NOTIFY
+#include <sys/vfs.h>
+#include <libnotify/notify.h>
+#endif
+
 #include <gnome.h>
 #include <glib/gi18n.h>
 #include <gconf/gconf-client.h>
@@ -65,6 +77,23 @@ static GSList *mounted_volumes = NULL;
 /* Table of active dialogs indexed by the udi of the device in question */
 static GHashTable *dialogs = NULL;
 
+/* List of UDIs of all mounted volumes being watched for low-disk notification */
+static GHashTable *statfs_mounts = NULL;
+
+/* Timeout for the callback to do statfs () on mounted volumes */
+static guint statfs_id = 0;
+
+
+typedef struct {
+        gboolean notified;
+        double last_pct;
+	char *udi;
+} statfs_mount_info;
+
+static void statfs_mount_info_add (const char *udi);
+static void statfs_mount_info_free (statfs_mount_info *info);
+static gboolean gvm_statfs_check_space (const char *udi, statfs_mount_info *info, gpointer user_data);
+
 
 static gboolean gvm_user_at_console (void);
 static gboolean gvm_user_active_at_console (void);
@@ -72,7 +101,8 @@ static gboolean gvm_user_active_at_console (void);
 
 typedef enum {
 	TYPE_BOOL,
-	TYPE_STRING
+	TYPE_STRING,
+	TYPE_FLOAT
 } type_t;
 
 enum {
@@ -100,6 +130,11 @@ enum {
 	AUTOPRINTER_COMMAND,
 	AUTORUN,
 	AUTORUN_PATH,
+	AUTOTABLET,
+	AUTOTABLET_COMMAND,
+	PERCENT_THRESHOLD,
+	PERCENT_FREED,
+	PERCENT_USED
 };
 
 static struct {
@@ -137,6 +172,9 @@ static struct {
 	{ GCONF_ROOT "autorun_path",              TYPE_STRING, &config.autorun_path              },
 	{ GCONF_ROOT "autotablet",                TYPE_BOOL,   &config.autotablet                },
 	{ GCONF_ROOT "autotablet_command",        TYPE_STRING, &config.autotablet_command        },
+	{ GCONF_ROOT "percent_threshold",         TYPE_FLOAT,  &config.percent_threshold         },
+	{ GCONF_ROOT "percent_freed",             TYPE_FLOAT,  &config.percent_freed             },
+	{ GCONF_ROOT "percent_used",              TYPE_FLOAT,  &config.percent_used              },
 };
 
 static GHashTable *gvm_settings_hash = NULL;
@@ -583,6 +621,15 @@ gvm_load_config (void)
 				gconf_client_get_bool (config.client, gvm_settings[i].key, NULL);
 			dbg ("setting[%d]: bool: %s = %d\n", i, strrchr (gvm_settings[i].key, '/') + 1,
 			     *((int *) gvm_settings[i].var));
+		} else if (gvm_settings[i].type == TYPE_FLOAT) {
+			*((double *) gvm_settings[i].var) =
+				gconf_client_get_float (config.client, gvm_settings[i].key, NULL);
+			if (*((double *) gvm_settings[i].var) >= 1.0)
+				*((double *) gvm_settings[i].var) = 1.0;
+			else if (*((double *) gvm_settings[i].var) <= 0.0)
+				*((double *) gvm_settings[i].var) = 0.0;
+			dbg ("settings[%d]: float: %s = %f\n", i, strrchr (gvm_settings[i].key, '/') + 1,
+			     *((double *) gvm_settings[i].var));
 		} else {
 			g_assert_not_reached ();
 		}
@@ -623,6 +670,14 @@ gvm_config_changed (GConfClient *client __attribute__((__unused__)),
 		*((int *) gvm_settings[which].var) = gconf_value_get_bool (value);
 		dbg ("setting changed: bool: %s = %d\n", strrchr (gvm_settings[which].key, '/') + 1,
 		     *((int *) gvm_settings[which].var));
+	} else if (gvm_settings[which].type == TYPE_FLOAT) {
+		*((double *) gvm_settings[which].var) = gconf_client_get_float (config.client, gvm_settings[which].key, NULL);
+		if (*((double *) gvm_settings[which].var) >= 1.0)
+			*((double *) gvm_settings[which].var) = 1.0;
+		else if (*((double *) gvm_settings[which].var) <= 0.0)
+			*((double *) gvm_settings[which].var) = 0.0;
+		dbg ("settings[%d]: float: %s = %f\n", which, strrchr (gvm_settings[which].key, '/') + 1,
+		     *((double *) gvm_settings[which].var));
 	} else {
 		g_assert_not_reached ();
 	}
@@ -1193,6 +1248,35 @@ gvm_udi_is_subfs_mount (const char *udi)
 	return subfs;
 }
 
+static gboolean
+gvm_storage_device_is_cdrom (const char *storage)
+{
+	gboolean is_cdrom = FALSE;
+	char *media_type;
+	
+	if ((media_type = libhal_device_get_property_string (hal_ctx, storage, "storage.drive_type", NULL))) {
+		is_cdrom = !strcmp (media_type, "cdrom");
+		libhal_free_string (media_type);
+	}
+	
+	return is_cdrom;
+}
+
+static gboolean
+gvm_udi_is_cdrom (const char *udi)
+{
+	gboolean is_cdrom;
+	char *storage;
+	
+	if (!(storage = libhal_device_get_property_string (hal_ctx, udi, "block.storage_device", NULL)))
+		return FALSE;
+	
+	is_cdrom = gvm_storage_device_is_cdrom (storage);
+	libhal_free_string (storage);
+	
+	return is_cdrom;
+}
+
 static void
 import_storage_camera_cb (GvmPromptCtx *ctx, int action)
 {
@@ -1287,6 +1371,184 @@ gvm_device_mounted (const char *udi)
 }
 
 
+enum {
+	MOUNT_CODEPAGE   = (1 << 0),
+	MOUNT_DATA       = (1 << 1),
+	MOUNT_DIRSYNC    = (1 << 2),
+	MOUNT_DMASK      = (1 << 3),
+	MOUNT_FLUSH      = (1 << 4),
+	MOUNT_IOCHARSET  = (1 << 5),
+	MOUNT_MODE       = (1 << 6),
+	MOUNT_NOATIME    = (1 << 7),
+	MOUNT_NODIRATIME = (1 << 8),
+	MOUNT_NOEXEC     = (1 << 9),
+	MOUNT_QUIET      = (1 << 10),
+	MOUNT_READ_ONLY  = (1 << 11),
+	MOUNT_SHORTNAME  = (1 << 12),
+	MOUNT_SYNC       = (1 << 13),
+	MOUNT_UID        = (1 << 14),
+	MOUNT_UMASK      = (1 << 15),
+	MOUNT_UTF8       = (1 << 16),
+};
+
+static struct {
+	const char *name;
+	guint32 flag;
+} mount_options[] = {
+	{ "codepage=",  MOUNT_CODEPAGE   },  /* vfat */
+	{ "data=",      MOUNT_DATA       },  /* ext3 */
+	{ "dirsync",    MOUNT_DIRSYNC    },
+	{ "dmask=",     MOUNT_DMASK      },  /* ntfs */
+	{ "flush",      MOUNT_FLUSH      },  /* vfat */
+	{ "iocharset=", MOUNT_IOCHARSET  },  /* vfat, iso9660 */
+	{ "mode=",      MOUNT_MODE       },  /* iso9660 */
+	{ "noatime",    MOUNT_NOATIME    },
+	{ "nodiratime", MOUNT_NODIRATIME },
+	{ "noexec",     MOUNT_NOEXEC     },
+	{ "quiet",      MOUNT_QUIET      },
+	{ "ro",         MOUNT_READ_ONLY  },
+	{ "shortname=", MOUNT_SHORTNAME  },  /* vfat */
+	{ "sync",       MOUNT_SYNC       },
+	{ "uid=",       MOUNT_UID        },  /* vfat, ntfs, udf, iso9660 */
+	{ "umask=",     MOUNT_UMASK      },  /* vfat, ntfs, udf */
+	{ "utf8",       MOUNT_UTF8       },  /* vfat, iso9660 */
+};
+
+
+static const char *
+canonicalize_codeset (char *str)
+{
+	register char *s = str;
+	register char *d = str;
+	
+	while (*s) {
+		if (*s >= 'A' && *s <= 'Z')
+			*d++ = *s++ + 0x20;
+		else if (*s != '-' && *s != '_')
+			*d++ = *s++;
+		else
+			s++;
+	}
+	
+	*d = '\0';
+	
+	return str;
+}
+
+static struct {
+	const char *codeset;
+	const char *iocharset;
+} iocharset_mapping[] = {
+	{ "tis620",      "cp874"      },
+	
+	{ "shiftjis*",   "cp932"      },
+	{ "sjis",        "cp932"      },
+	
+	{ "gb18030",     "cp936"      },
+	{ "gbk",         "cp936"      },
+	{ "gb2312",      "cp936"      },
+	
+	{ "euckr",       "cp949"      },
+	
+	{ "big5*",       "cp950"      },
+	{ "euctw",       "cp950"      },
+	
+	/*{ "cp1251",      "cp1251"     },*/
+	
+	{ "eucjp",       "euc-jp"     },
+	
+	{ "iso88591",    "iso8859-1"  },
+	{ "iso88592",    "iso8859-2"  },
+	{ "iso88593",    "iso8859-3"  },
+	{ "iso88594",    "iso8859-4"  },
+	{ "iso88595",    "iso8859-5"  },
+	{ "iso88596",    "iso8859-6"  },
+	{ "iso88597",    "iso8859-7"  },
+	{ "iso88599",    "iso8859-9"  },
+	{ "iso885913",   "iso8859-13" },
+	{ "iso885914",   "iso8859-14" },
+	{ "iso885915",   "iso8859-15" },
+	
+	{ "koi8r",       "koi8-r"     },
+	{ "koi8u",       "koi8-u"     },
+	
+	/*{ "utf8",       "utf8"       },*/
+};
+
+static const char *
+gvm_iocharset (void)
+{
+	static const char *iocharset = NULL;
+	char *locale, *codeset, *inptr;
+	static int initialized = FALSE;
+	const char *wildcard;
+	size_t i, n;
+	
+	if (initialized)
+		return iocharset;
+	
+	initialized = TRUE;
+	
+	locale = setlocale (LC_ALL, NULL);
+	if (!locale || !strcmp (locale, "C") || !strcmp (locale, "POSIX")) {
+		/* The locale "C"  or  "POSIX"  is  a  portable  locale;  its
+		 * LC_CTYPE  part  corresponds  to  the 7-bit ASCII character
+		 * set.
+		 */
+		
+		iocharset = NULL;
+	} else {
+#ifdef HAVE_CODESET
+		codeset = g_strdup (nl_langinfo (CODESET));
+		canonicalize_codeset (codeset);
+#else
+		/* A locale name is typically of  the  form  language[_terri-
+		 * tory][.codeset][@modifier],  where  language is an ISO 639
+		 * language code, territory is an ISO 3166 country code,  and
+		 * codeset  is  a  character  set or encoding identifier like
+		 * ISO-8859-1 or UTF-8.
+		 */
+		if (!(codeset = strchr (locale, '.'))) {
+			/* charset unknown */
+			return NULL;
+		}
+		
+		codeset++;
+		
+		/* ; is a hack for debian systems and / is a hack for Solaris systems */
+		inptr = codeset;
+		while (*inptr && !strchr ("@;/", *inptr))
+			inptr++;
+		
+		codeset = g_strndup (codeset, inptr - codeset);
+		canonicalize_codeset (codeset);
+#endif
+		
+		for (i = 0; i < G_N_ELEMENTS (iocharset_mapping); i++) {
+			if ((wildcard = strchr (iocharset_mapping[i].codeset, '*'))) {
+				n = wildcard - iocharset_mapping[i].codeset;
+				if (!strncmp (codeset, iocharset_mapping[i].codeset, n)) {
+					iocharset = iocharset_mapping[i].iocharset;
+					break;
+				}
+			} else if (!strcmp (codeset, iocharset_mapping[i].codeset)) {
+				iocharset = iocharset_mapping[i].iocharset;
+				break;
+			}
+		}
+		
+		if (!iocharset) {
+			iocharset = codeset;
+			codeset = NULL;
+		}
+		
+		g_free (codeset);
+	}
+	
+	return iocharset;
+}
+
+
 /*
  * gvm_device_mount - mount the given device.
  *
@@ -1331,12 +1593,10 @@ gvm_device_mount (const char *udi, gboolean interactive)
 		
 		return retval;
 	} else {
-		char *mount_point, *fstype;
+		char *mount_point, *fstype, *charset_opt = NULL;
 		DBusMessage *dmesg, *reply;
-		dbus_uint64_t size;
 		GPtrArray *options;
 		DBusError error;
-		char uid[32];
 		
 		if (!(dmesg = dbus_message_new_method_call ("org.freedesktop.Hal", udi,
 							    "org.freedesktop.Hal.Device.Volume",
@@ -1346,19 +1606,59 @@ gvm_device_mount (const char *udi, gboolean interactive)
 		}
 		
 		options = g_ptr_array_new ();
-		size = libhal_device_get_property_uint64 (hal_ctx, udi, "volume.size", NULL);
 		if ((fstype = libhal_device_get_property_string (hal_ctx, udi, "volume.fstype", NULL))) {
-			snprintf (uid, sizeof (uid) - 1, "uid=%u", getuid ());
+			char **moptions = NULL;
+			const char *iocharset;
+			guint32 opts = 0;
+			char uid[32];
+			size_t i, j;
+			
+			if ((moptions = libhal_device_get_property_strlist (hal_ctx, udi, "volume.mount.valid_options", NULL))) {
+				for (i = 0; moptions[i]; i++) {
+					for (j = 0; j < G_N_ELEMENTS (mount_options); j++) {
+						if (!strcmp (moptions[i], mount_options[j].name))
+							opts |= mount_options[j].flag;
+					}
+				}
+				
+				libhal_free_string_array (moptions);
+			}
 			
 			if (!strcmp (fstype, "vfat")) {
-				g_ptr_array_add (options, "shortname=winnt");
-				g_ptr_array_add (options, uid);
+				/* preferably mount flush (opportunistic syncing) if available, else
+				 * mount sync for "small" volumes (that are likely to be thumb drives)
+				 */
+				if (opts & MOUNT_FLUSH) {
+					g_ptr_array_add (options, "flush");
+				} else if (opts & MOUNT_SYNC) {
+					dbus_uint64_t size;
+					
+					size = libhal_device_get_property_uint64 (hal_ctx, udi, "volume.size", NULL);
+					if (size <= (512 * 1024 * 1024))
+						g_ptr_array_add (options, "sync");
+				}
 				
-				if (size <= (512 * 1024 * 1024))
-					g_ptr_array_add (options, "sync");
+				if (opts & MOUNT_SHORTNAME)
+					g_ptr_array_add (options, "shortname=winnt");
 			} else if (!strcmp (fstype, "iso9660")) {
-				g_ptr_array_add (options, uid);
+				/* only care about uid= and iocharset= */
 			} else if (!strcmp (fstype, "udf")) {
+				/* only care about uid= and iocharset= */
+			}
+			
+			if (opts & (MOUNT_IOCHARSET|MOUNT_UTF8)) {
+				if ((iocharset = gvm_iocharset ())) {
+					if ((opts & MOUNT_UTF8) && !strcmp (iocharset, "utf8")) {
+						g_ptr_array_add (options, "utf8");
+					} else if (opts & MOUNT_IOCHARSET) {
+						charset_opt = g_strdup_printf ("iocharset=%s", iocharset);
+						g_ptr_array_add (options, charset_opt);
+					}
+				}
+			}
+			
+			if (opts & MOUNT_UID) {
+				snprintf (uid, sizeof (uid) - 1, "uid=%u", getuid ());
 				g_ptr_array_add (options, uid);
 			}
 			
@@ -1377,6 +1677,7 @@ gvm_device_mount (const char *udi, gboolean interactive)
 		}
 		
 		g_ptr_array_free (options, TRUE);
+		g_free (charset_opt);
 		
 		policy = g_new (struct _MountPolicy, 1);
 		policy->udi = g_strdup (udi);
@@ -1678,7 +1979,6 @@ static gboolean
 gvm_media_changed (const char *udi, const char *storage_device)
 {
 	gboolean handled = FALSE;
-	char *media_type;
 	DBusError error;
 	
 	/* Refuse to enforce policy on removable media if drive is locked */
@@ -1690,26 +1990,11 @@ gvm_media_changed (const char *udi, const char *storage_device)
 		return TRUE;
 	}
 	
-	/*
-	 * Get HAL's interpretation of our media type.  Note that we must check
-	 * the storage device and not this UDI
-	 */
-	if (!(media_type = libhal_device_get_property_string (hal_ctx, storage_device, "storage.drive_type", &error))) {
-		warn ("cannot get storage.drive_type: %s", error.message);
-		dbus_error_free (&error);
-		
-		/* FIXME: should we really return TRUE here? not sure... need a test case */
-		
-		return TRUE;
-	}
-	
-	if (!strcmp (media_type, "cdrom")) {
+	if (gvm_storage_device_is_cdrom (storage_device)) {
 		gvm_cdrom_policy (udi);
 		handled = TRUE;
 	}
 	
-	libhal_free_string (media_type);
-	
 	return handled;
 }
 
@@ -2104,9 +2389,29 @@ hal_property_modified (LibHalContext *ctx __attribute__((__unused__)),
 			dbg ("not in mount queue: %s\n", udi);
 			gvm_device_mounted (udi);
 		}
+		
+#ifdef ENABLE_NOTIFY
+		/* If we aren't a cdrom, then add to the statfs list */
+		if (!gvm_udi_is_cdrom (udi)) {
+			if (!g_hash_table_lookup (statfs_mounts, udi))
+				statfs_mount_info_add (udi);
+		}
+#endif
 	} else {
 		dbg ("Unmounted: %s\n", udi);
 		
+#ifdef ENABLE_NOTIFY
+		/* remove the udi from the statfs_mounts list */
+		if (!gvm_udi_is_cdrom (udi)) {
+			statfs_mount_info *info;
+			
+                        if ((info = g_hash_table_lookup (statfs_mounts, udi))) {
+				g_hash_table_remove (statfs_mounts, udi);
+				statfs_mount_info_free (info);
+			}
+		}
+#endif
+		
 		/* unmount all volumes mounted during lifetime */
 		
 		for (l = mounted_volumes; l != NULL; l = n) {
@@ -2332,14 +2637,30 @@ mount_all (LibHalContext *ctx)
 	
 	for (i = 0; i < num_volumes; i++) {
 		udi = volumes[i];
-		
-		if (gvm_udi_is_subfs_mount (udi))
+
+		if (gvm_udi_is_subfs_mount (udi)) {
+#ifdef ENABLE_NOTIFY
+			/* Add to the statfs_mounts list if not a cdrom */
+			if (!gvm_udi_is_cdrom (udi)) {
+				if (!g_hash_table_lookup (statfs_mounts, udi))
+					statfs_mount_info_add (udi);
+			}
+#endif
 			continue;
+		}
 		
 		/* don't attempt to mount already mounted volumes */
 		if (!libhal_device_property_exists (ctx, udi, "volume.is_mounted", NULL)
-		    || libhal_device_get_property_bool (ctx, udi, "volume.is_mounted", NULL))
+		    || libhal_device_get_property_bool (ctx, udi, "volume.is_mounted", NULL)) {
+#ifdef ENABLE_NOTIFY
+			/* Add to the statfs_mounts list if not a cdrom */
+			if (!gvm_udi_is_cdrom (udi)) {
+                                if (!g_hash_table_lookup (statfs_mounts, udi))
+					statfs_mount_info_add (udi);
+			}
+#endif
 			continue;
+		}
 		
 		/* only mount if the block device has a sensible filesystem */
 		if (!libhal_device_property_exists (ctx, udi, "volume.fsusage", NULL))
@@ -2442,12 +2763,116 @@ sigterm_iochn_data (GIOChannel *source,
 	
 	unmount_all ();
 	
+#ifdef ENABLE_NOTIFY
+	g_source_remove (statfs_id);
+	statfs_id = 0;
+	
+	g_hash_table_destroy (statfs_mounts);
+#endif
+
 	gtk_main_quit ();
 	
  out:
 	return TRUE;
 }
 
+
+static void
+statfs_mount_info_add (const char *udi)
+{
+	statfs_mount_info *info;
+	
+	info = g_new0 (statfs_mount_info, 1);
+	info->udi = g_strdup (udi);
+	info->notified = FALSE;
+	info->last_pct = 0.0;
+	
+	g_hash_table_insert (statfs_mounts, info->udi, info);
+	
+	gvm_statfs_check_space (info->udi, info, NULL);
+}
+
+static void
+statfs_mount_info_free (statfs_mount_info *info)
+{
+	g_free (info->udi);
+	g_free (info);
+}
+
+static gboolean
+gvm_statfs_check_space (const char *udi, statfs_mount_info *info, gpointer user_data)
+{
+#ifdef ENABLE_NOTIFY
+	char *mount_point = NULL;
+	NotifyNotification *n;
+	struct statfs buf;
+	
+	if (!(mount_point = libhal_device_get_property_string (hal_ctx, udi, "volume.mount_point", NULL)))
+		return TRUE;
+	
+	if (statfs (mount_point, &buf) != -1) {
+		double free_space;
+		
+		free_space = (double) buf.f_bavail / (double) buf.f_blocks;
+		
+		if ((free_space - info->last_pct) > config.percent_used)
+			info->notified = FALSE;
+		else if ((free_space - info->last_pct) < config.percent_freed)
+			info->notified = TRUE;
+		
+		if (!info->notified && free_space < config.percent_threshold) {
+			char *low_disk_msg = NULL;
+			char *label = NULL;
+			int free_pct;
+			
+			label = libhal_device_get_property_string (hal_ctx, udi, "volume.label", NULL);
+			if (!label || label[0] == '\0') {
+				libhal_free_string (label);
+				label = g_strdup (mount_point);
+			}
+			
+			free_pct = 100 - (free_space * 100);
+			low_disk_msg = g_strdup_printf (_("%d%% of the disk space on %s is in use"), free_pct, label);
+			
+			n = notify_notification_new ("Low Disk Space", low_disk_msg, "gnome-dev-harddisk", NULL);
+			notify_notification_set_urgency (n, NOTIFY_URGENCY_CRITICAL);
+			notify_notification_show (n, NULL);
+			
+			g_object_unref (n);
+			
+			libhal_free_string (low_disk_msg);
+			libhal_free_string (label);
+			
+			info->notified = TRUE;
+		}
+		
+		info->last_pct = free_space;
+	}
+	
+	libhal_free_string (mount_point);
+	
+	return TRUE;
+#else
+	udi = udi;
+	info = info;
+	user_data = user_data;
+	
+	return FALSE;
+#endif
+}
+
+static gboolean
+gvm_statfs_timeout (gpointer data)
+{
+#ifdef ENABLE_NOTIFY
+	g_hash_table_foreach (statfs_mounts, (GHFunc) gvm_statfs_check_space, NULL);
+	
+	return TRUE;
+#else
+	return FALSE;
+#endif
+}
+
 static void
 gvm_die (GnomeClient *client __attribute__((__unused__)),
 	 gpointer user_data __attribute__((__unused__)))
@@ -2455,6 +2880,13 @@ gvm_die (GnomeClient *client __attribute__((__unused__)),
 	dbg ("Received 'die', initiating shutdown\n");
 	
 	unmount_all ();
+
+#ifdef ENABLE_NOTIFY
+	g_source_remove (statfs_id);
+	statfs_id = 0;
+	
+	g_hash_table_destroy (statfs_mounts);
+#endif
 	
 	gtk_main_quit ();
 }
@@ -2580,10 +3012,25 @@ main (int argc, char **argv)
 	mount_table = g_hash_table_new (g_str_hash, g_str_equal);
 	dialogs = g_hash_table_new (g_str_hash, g_str_equal);
 	
+#ifdef ENABLE_NOTIFY
+	/* Initialize the notification connection */
+	notify_init ("GNOME Volume Manager");
+	
+	/* Initialize the statfs_mounts hash table */
+	statfs_mounts = g_hash_table_new (g_str_hash, g_str_equal);
+#endif
+
 	mount_all (hal_ctx);
+
+	/* Add a timeout for our statfs checking */
+	statfs_id = g_timeout_add (15000, (GSourceFunc) gvm_statfs_timeout, NULL);
 	
 	gtk_main ();
 	
+#ifdef ENABLE_NOTIFY
+	g_hash_table_destroy (statfs_mounts);
+#endif
+	
 	g_hash_table_destroy (mount_table);
 	g_object_unref (program);
 	
